library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity ups_inout is port(
  sys_clk : in std_logic;
  reset : IN std_logic;
  ups_valid : in std_logic;  -- control from PS
  BRAM0_addr : OUT std_logic_vector(31 downto 0);
  BRAM0_data : IN std_logic_vector(31 downto 0);
  BRAM1_addr : OUT std_logic_vector(31 downto 0);
  BRAM1_data : IN std_logic_vector(31 downto 0);
  Flip : OUT std_logic_vector(7 downto 0);

  BRAM2_addr : OUT std_logic_vector(31 downto 0);
  BRAM2_data : OUT std_logic_vector(31 downto 0);
  BRAM2_we : OUT std_logic_vector(3 downto 0);
  BRAM3_addr : OUT std_logic_vector(31 downto 0);
  BRAM3_data : OUT std_logic_vector(31 downto 0);
  BRAM3_we : OUT std_logic_vector(3 downto 0);
  Flip1 : OUT std_logic_vector(7 downto 0)
);
end entity ups_inout;

architecture behavioral of ups_inout is
 
  COMPONENT upsampler PORT (
    clock : IN STD_LOGIC;
    r_ext : IN STD_LOGIC_VECTOR(15 DOWNTO 0);
    d0, d1, d2, d3 : OUT STD_LOGIC_VECTOR(15 DOWNTO 0)
  );
  END COMPONENT;  

  signal addrin : std_logic_vector(16 downto 0) := (others => '0'); -- 131072
  signal addrout : std_logic_vector(17 downto 0) := (others => '0'); -- 262144
  signal we2 : std_logic_vector(3 downto 0) := (others => '0');
  signal we3 : std_logic_vector(3 downto 0) := (others => '0');

  signal bit_counter : std_logic_vector(6 downto 0) := (others => '0');  -- empirical size for M4, 64 taps
  signal valid : std_logic := '0';
  signal sample : std_logic_vector(31 downto 0) := (others => '0');  -- left & right signed 16-bit sample

  signal upsampler_clock : std_logic;
  signal d0L, d1L, d2L, d3L : std_logic_vector(15 downto 0) := (others => '0');
  signal d0R, d1R, d2R, d3R : std_logic_vector(15 downto 0) := (others => '0');

begin
  
  -- 32768 * 4 in 16 times 2048 * 4 steps
  BRAM0_addr <= x"0000" & "000" & addrin(12 downto 0); -- 2048*4=8192
  BRAM1_addr <= x"0000" & "000" & addrin(12 downto 0); 
  Flip <= "0000" & addrin(16 downto 13);  -- indicates BRAM0/BRAM1 flip to PS

  BRAM2_addr <= x"0000" & "00" & addrout(13 downto 0); -- 4096*4=16384
  BRAM3_addr <= x"0000" & "00" & addrout(13 downto 0); 
  Flip1 <= "0000" & addrout(17 downto 14);  -- indicates BRAM2/BRAM3 flip to PS

  BRAM2_we <= we2;
  BRAM3_we <= we3;

  Inst_upsampler_left: upsampler PORT MAP (
    clock => upsampler_clock,
    r_ext => sample(15 downto 0) xor x"8000",
    d0 => d0L,
    d1 => d1L,
    d2 => d2L,
    d3 => d3L
  );

  Inst_upsampler_right: upsampler PORT MAP (
    clock => upsampler_clock,
    r_ext => sample(31 downto 16) xor x"8000",
    d0 => d0R,
    d1 => d1R,
    d2 => d2R,
    d3 => d3R
  );

  process (sys_clk)
  begin
    if sys_clk'event and sys_clk = '1' then
      if reset = '1' then 
        valid <= '0';
        bit_counter <= (others => '0');
        addrin <= (others => '0');
        addrout <= (others => '0');
      else      
        upsampler_clock <= '0';
        we2 <= "0000";
        we3 <= "0000";
      
        if ups_valid = '1' and valid = '0' then
          valid <= '1';
          bit_counter <= (others => '0');
        end if;

        if ups_valid = '0' and valid = '1' then
          case bit_counter is 
          when "0000000" =>
            if addrin(13) = '0' then
              sample <= BRAM0_data;
            else
              sample <= BRAM1_data;
            end if;
            
          when "0000001" =>
            upsampler_clock <= '1';

          when "1111011" =>
            if addrout(14) = '0' then 
              BRAM2_data <= (d1R xor x"8000") & (d1L xor x"8000");
              we2 <= "1111";
            else
              BRAM3_data <= (d1R xor x"8000") & (d1L xor x"8000");
              we3 <= "1111";
            end if;
            
          when "1111100" =>
            addrout <= addrout+4;  -- next stereo sample
       
          when "1111101" =>
            if addrout(14) = '0' then 
              BRAM2_data <= (d3R xor x"8000") & (d3L xor x"8000");
              we2 <= "1111";
            else
              BRAM3_data <= (d3R xor x"8000") & (d3L xor x"8000");
              we3 <= "1111";
            end if;
            
          when "1111110" =>
            addrin <= addrin+4;  -- next stereo sample
            addrout <= addrout+4;

          when "1111111" =>
            if addrin = "0000000000000" then
              valid <= '0';
            end if;
        
          when others => null; 
          end case;
          bit_counter <= bit_counter + 1;
        end if;  
      end if;
    end if;
  end process;
  
end behavioral;
